;; The first three lines of this file were inserted by DrScheme. They record metadata
;; about the language level of this file in a form that our tools can easily process.
#reader(lib "reader.ss" "plai" "lang")
(load "env.scm")
(load "parser.scm")

(define-struct expression ())
(define-struct (lit-exp   expression) (value))
(define-struct (var-exp   expression) (name))
(define-struct (if-exp    expression) (test then else))
(define-struct (begin-exp expression) (exps))
(define-struct (let-exp   expression) (params args body))
(define-struct (fun-exp   expression) (params body))
(define-struct (app-exp   expression) (fun args))
(define-struct (prim-exp  expression) (prim args))


;; reflective access to current env
(define-struct (env-capt-exp  expression) ())

;; reflective access to current jp
(define-struct (this-jp-exp  expression) ())
        
;; closure values
(define-struct value ())
(define-struct (closure value) (params body env))


;; scoped let
(define-struct (slet-exp  expression) (ds params args body))

;; accessors for deployment strategy components (ds is a list)
(define (c ds)
  (car ds))

(define (d ds)
  (cadr ds))

(define (f ds)
  (caddr ds))


;; join point stack
(define-struct jp   (closure args parent))

(define root-jp #f)

(define (jp-depth jp)
  (if (not jp)
      0
      (if (not (jp-parent jp))
          1
          (+ 1 (jp-depth (jp-parent jp))))))

(define (stack-depth-less n jp)
  (lambda (jp2)
    (<= n (- (jp-depth jp2) (jp-depth jp)))))

(define (run pgm)
  (interp (parse pgm) (make-empty-env) root-jp))

;; evaluate expression exp in environment env, with current join point jp (optional)
(define (interp exp env jp)
  (cond ((lit-exp? exp) (lit-exp-value exp))
        
        ((prim-exp? exp)
         (let ((args (interp-args (prim-exp-args exp) env jp)))
           (apply (prim-exp-prim exp) args)))
        
        ((if-exp? exp)
         (if (interp (if-exp-test exp) env jp)
             (interp (if-exp-then exp) env jp)
             (interp (if-exp-else exp) env jp)))
        
        ((begin-exp? exp)
         (interp-body (begin-exp-exps exp) env jp))
        
        ((slet-exp? exp)
         (let ((ds   (interp (slet-exp-ds exp) env jp))
               (args (interp-args (slet-exp-args exp) env jp)))
           (interp-body (slet-exp-body exp) (extend-env (slet-exp-params exp) args env ds 'l) jp)))
                                                                                           
        ((var-exp? exp)   (env-get (var-exp-name exp) env (match-f jp)))
        
        ((fun-exp? exp)   (make-closure (fun-exp-params exp) (fun-exp-body exp) (collect-match-d jp env))) ;; could create def jp
        
        ((app-exp? exp)   (let* ((cl     (interp (app-exp-fun exp) env jp))
                                 (args   (interp-args (app-exp-args exp) env jp)))
                            (apply-cl cl args env jp)))
                                 
        ((this-jp-exp? exp) jp)
        ((env-capt-exp? exp) env)
        ))

(define (interp-args exps env jp)
    (map (lambda (x) (interp x env jp)) exps))

(define (interp-body body env jp)
  (let loop ((exps body) (val #f))
    (if (null? exps)
        val
        (loop (cdr exps) (interp (car exps) env jp)))))

;;applies the function f, passing jp and env as argument
;; only bindings captured in the closure are available
(define (app f jp env)
  (apply-cl f (list jp env) (make-empty-env) jp))

(define (apply-cl cl args env jp)
  (let* ((njp    (make-jp cl args jp))
         (clenv  (extend-env (closure-params cl) args (closure-env cl) ds-static 'a)) ;; fun args are lexically scoped
                                                                                   ; (can easily support dyn-scoped args of λd)
         (denv   (collect-match-c njp env)))
    (interp-body (closure-body cl) (env-append clenv denv) njp))) ;; order preserves 'lexical scope reasoning'
                          

;; matching strategies
(define (match-component getter) 
  (lambda (jp)
    (lambda (e) (app (getter (extended-env-ds e)) jp e))))
  
(define (match-c jp) ((match-component c) jp))
(define (match-d jp) ((match-component d) jp))
(define (match-f jp) ((match-component f) jp))

;; collecting in environments
(define (collect-match-c jp env) ((collect-env-if 'c) (match-c jp) env))
(define (collect-match-d jp env) ((collect-env-if 'd) (match-d jp) env))
;(define (collect-match-f jp env) ((collect-env-if #f) (match-f jp) env)) 


;;;;;;;;;;;;;;;;;
;; strategies

(define f-true  (interp (parse '(fun (jp) #t)) (make-empty-env) root-jp))
(define f-false (interp (parse '(fun (jp) #f)) (make-empty-env) root-jp))

(define ds-static
  (list f-false f-true f-true))

(define ds-dynamic
  (list f-true f-false f-true))

(define ds-all
  (list f-true f-true f-true))

(define ds-flat
  (list f-false f-false f-true))

;; keeps propagation function, makes filter for given id: 
;; binding of a closure does not hide fluid binding
(define (make-ds-soft ds id)
  (list (c ds) (d ds) (interp (parse `(fun (jp env)
                                           (not (and (eq? 'd (env-mark env))
                                                     (env-check-mark (quote ,id) (env-next env) 'c)))))
                              (make-empty-env) root-jp)))